"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import pandas as pd

from ._localutil import fill_metrics_df, get_terminal_perf
from .core import Metric


class TerminalPerformance(Metric):
    name = "Terminal Performance"
    capability = "continual_learning"
    requires = {"syllabus_type": "agent"}
    description = "Calculates the terminal performance within each block"

    def __init__(self, perf_measure: str) -> None:
        super().__init__()
        self.perf_measure = perf_measure

    def validate(self, block_info) -> None:
        pass

    def calculate(
        self,
        dataframe: pd.DataFrame,
        block_info: pd.DataFrame,
        metrics_df: pd.DataFrame,
    ) -> pd.DataFrame:
        # Initialize metric dictionaries
        term_perf_values = {}
        exp_to_terminal_perf = {}

        # Iterate over all of the blocks and compute the within block performance
        for idx in range(max(block_info["regime_num"].to_numpy()) + 1):
            # Need to get the part of the data corresponding to the block
            block_data = dataframe.loc[dataframe["regime_num"] == idx]

            # Make within block calculations
            term_perf_values[idx], exp_to_terminal_perf[idx], _ = get_terminal_perf(
                block_data, col_to_use=self.perf_measure
            )

        metrics_df = fill_metrics_df(term_perf_values, "term_perf", metrics_df)
        return fill_metrics_df(exp_to_terminal_perf, "exp_to_term_perf", metrics_df)



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from typing import List, Tuple, Union

import numpy as np
import pandas as pd


def smooth(x: np.ndarray, window_len: int = None, window: str = "flat") -> np.ndarray:
    """Smooths the data using a window with requested size.

    Code from https://scipy-cookbook.readthedocs.io/items/SignalSmooth.html
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the beginning and end part of the output signal.

    example:
        t = linspace(-2,2,0.1)
        x = sin(t)+randn(len(t))*0.1
        y = smooth(x)

    See also:
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    NOTE: length(output) != length(input).
    To correct this, return y[(window_len/2-1):-(window_len/2)] instead of just y.

    Args:
        x (np.ndarray): The input signal.
        window_len (int, optional): The dimension of the smoothing window. Defaults to None.
        window (str, optional): The type of window from 'flat', 'hanning', 'hamming', 'bartlett',
            'blackman'. Flat window will produce a moving average smoothing. Defaults to 'flat'.

    Raises:
        ValueError: If input signal has more than one dimension.
        ValueError: If window type is not supported.

    Returns:
        np.ndarray: The smoothed signal.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if window_len is None or x.size < window_len:
        # raise(ValueError, "Input vector needs to be bigger than window size.")
        window_len = min(int(x.size * 0.2), 100)

    if window_len < 3:
        return x

    if window not in ["flat", "hanning", "hamming", "bartlett", "blackman"]:
        raise ValueError(
            "Window is one of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"
        )

    # Perform reflections at both ends of signal
    s = np.r_[x[window_len - 1 : 0 : -1], x, x[-2 : -window_len - 1 : -1]]
    s = s[~np.isnan(s)]

    if window == "flat":  # moving average
        w = np.ones(window_len, "d")
    else:
        w = eval("np." + window + "(window_len)")

    y = np.convolve(w / w.sum(), s, mode="valid")

    # Changed to return output of same length as input
    start_ind = int(np.floor(window_len / 2 - 1))
    end_ind = -int(np.ceil(window_len / 2))
    return y[start_ind:end_ind]


def get_block_saturation_perf(
    data: Union[pd.DataFrame, List],
    col_to_use: str = None,
    prev_sat_val: float = None,
    window_len: int = None,
) -> Tuple[float, int, int]:
    """Calculates the saturation value, experiences to saturation, and experiences to recovery.

    Args:
        data (Union[pd.DataFrame, List]): The input data.
        col_to_use (str): The column name of the metric to use for calculations. Defaults to None.
        prev_sat_val (float, optional): Previous saturation value for calculating recovery time.
            Defaults to None.
        window_len (int, optional): The window length for smoothing the data. Defaults to None.

    Returns:
        Tuple[float, int, int]: Saturation value, experiences to saturation, and experiences to recovery.
    """

    # Aggregate multiple reward values for the same experience
    if isinstance(data, pd.DataFrame):
        if data.exp_num.is_unique:
            mean_data = data[col_to_use].to_numpy()
        else:
            mean_reward_per_experience = (
                data.loc[:, ["exp_num", col_to_use]].groupby("exp_num").mean()
            )
            mean_data = np.ravel(mean_reward_per_experience.to_numpy())
    else:
        mean_data = np.array(data)

    # Take the moving average of the mean of the per experience reward
    smoothed_data = smooth(mean_data, window_len=window_len)
    smoothed_data = smoothed_data[~np.isnan(smoothed_data)]

    if len(smoothed_data):
        sat_val = np.max(smoothed_data)

        # Calculate the number of experiences to "saturation", which we define as the max of the moving average
        indices = np.where(smoothed_data == sat_val)
        exp_to_sat = int(indices[0][0])
        exp_to_rec = len(data) + 1

        if prev_sat_val:
            indices = np.where(smoothed_data >= prev_sat_val)
            if len(indices[0]):
                exp_to_rec = indices[0][0]
    else:
        sat_val = np.nan
        exp_to_sat = np.nan
        exp_to_rec = np.nan

    return sat_val, exp_to_sat, exp_to_rec


def get_terminal_perf(
    data: pd.DataFrame,
    col_to_use: str,
    prev_val: float = None,
    term_window_ratio: float = 0.1,
) -> Tuple[float, int, int]:
    """Calculates the terminal performance, experiences to terminal performance, and experiences to recovery.

    Args:
        data (pd.DataFrame): The input data.
        col_to_use (str): The column name of the metric to use for calculations.
        prev_val (float, optional): Previous saturation value for calculating recovery time.
            Defaults to None.
        term_window_ratio (float, optional): The ratio of terminal data points used to compute the
            terminal performance. Defaults to 0.1 for training blocks and 1.0 for evaluation blocks.

    Returns:
        Tuple[float, int, int]: Terminal performance, experiences to terminal performance,
            and experiences to recovery.
    """

    # Aggregate multiple reward values for the same experience
    if data.shape[0] > 1:
        if data.exp_num.is_unique:
            mean_data = data[col_to_use].to_numpy()
        else:
            mean_reward_per_experience = (
                data.loc[:, ["exp_num", col_to_use]].groupby("exp_num").mean()
            )
            mean_data = np.ravel(mean_reward_per_experience.to_numpy())
    else:
        mean_data = np.ravel(data[col_to_use].to_numpy(dtype=float))

    mean_data = mean_data[~np.isnan(mean_data)]

    if len(mean_data):
        # Average all data points in test block for terminal performance
        if data.iloc[0]["block_type"] == "test":
            term_window_ratio = 1

        terminal_value = np.mean(
            mean_data[int((1 - term_window_ratio) * mean_data.size) :]
        )

        # Calculate the number of experiences to terminal performance
        experiences_to_terminal_perf = int(
            (1 - (term_window_ratio / 2)) * mean_data.size
        )

        # Initialize recovery time to one more than number of learning experiences in the data
        experiences_to_recovery = len(data) + 1

        if prev_val is not None:
            indices = np.where(mean_data >= prev_val)
            if len(indices[0]):
                experiences_to_recovery = indices[0][0]
    else:
        terminal_value = np.nan
        experiences_to_terminal_perf = np.nan
        experiences_to_recovery = np.nan

    return terminal_value, experiences_to_terminal_perf, experiences_to_recovery


def fill_metrics_df(
    data: dict, metric_string_name: str, metrics_df: pd.DataFrame, dict_key: str = None
) -> pd.DataFrame:
    """Fills the metrics DataFrame with additional data.

    Args:
        data (dict): The new metric data to insert into the metrics DataFrame.
        metric_string_name (str): The name of the new metric to add.
        metrics_df (pd.DataFrame): The metrics DataFrame to insert the new data into.
        dict_key (str, optional): The dictionary key of the metrics DataFrame to insert the new
            metric into, allows a higher level of insertion if value is passed. Defaults to None.

    Returns:
        pd.DataFrame: The updated metrics DataFrame.
    """

    if not dict_key:
        metrics_df[metric_string_name] = np.full_like(
            metrics_df["regime_num"], np.nan, dtype=np.object
        )
        for idx in data.keys():
            metrics_df.at[idx, metric_string_name] = data[idx]
    else:
        metrics_df[dict_key][metric_string_name] = np.full_like(
            metrics_df[dict_key]["regime_num"], np.nan, dtype=np.double
        )
        for idx in data.keys():
            metrics_df[dict_key].loc[idx, metric_string_name] = data[idx]

    return metrics_df


def get_simple_rl_task_names(task_names: list) -> list:
    """Simplifies the task name.

    For each task name in the provided list, this function splits the names using an underscore as
    delimiter, then returns the last element in the split list as the simplified name.

    Args:
        task_names (list): The list of task names to simplify.

    Returns:
        list: The list of simplified task names.
    """

    simple_names = []

    for t in task_names:
        splits = str.split(t, "_")
        simple_names.append(splits[-1].lower())

    return simple_names



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import argparse


def init_parser():
    # Instantiate parser
    parser = argparse.ArgumentParser(description="Run L2Metrics from the command line")

    # Log directories can be absolute paths, relative paths, or paths found in $L2DATA/logs
    parser.add_argument(
        "-l",
        "--log-dir",
        default=None,
        type=str,
        help="Log directory of scenario. Defaults to None.",
    )

    # Flag for recursively calculating metrics on valid subdirectories within log directory
    parser.add_argument(
        "-R",
        "--recursive",
        action="store_true",
        help="Recursively compute metrics on logs found in specified directory. \
                            Defaults to false.",
    )

    # Method for handling task variants
    parser.add_argument(
        "-r",
        "--variant-mode",
        default="aware",
        type=str,
        choices=["aware", "agnostic"],
        help="Mode for computing metrics with respect to task variants. \
                            Defaults to aware.",
    )

    # Mode for storing log data as STE data
    parser.add_argument(
        "-s",
        "--ste-store-mode",
        default=None,
        choices=["w", "a"],
        help="Mode for storing log data as STE, overwrite (w) or append (a). \
                            Defaults to None.",
    )

    # Method for handling multiple STE runs
    parser.add_argument(
        "-v",
        "--ste-averaging-method",
        default="metrics",
        choices=["metrics", "time"],
        help="Method for handling STE runs, LL metric averaging (metrics) or "
        "time-series averaging (time). Defaults to metrics.",
    )

    # Choose application measure to use as performance column
    parser.add_argument(
        "-p",
        "--perf-measure",
        default="reward",
        type=str,
        help="Name of column to use for metrics calculations. Defaults to reward.",
    )

    # Method for aggregating within-lifetime metrics
    parser.add_argument(
        "-a",
        "--aggregation-method",
        default="mean",
        type=str,
        choices=["mean", "median"],
        help="Method for aggregating within-lifetime metrics. Defaults to mean.",
    )

    # Method for calculating performance maintenance
    parser.add_argument(
        "-m",
        "--maintenance-method",
        default="mrlep",
        type=str,
        choices=["mrlep", "mrtlp", "both"],
        help="Method for computing performance maintenance. Defaults to mrlep.",
    )

    # Method for calculating forward and backward transfer
    parser.add_argument(
        "-t",
        "--transfer-method",
        default="ratio",
        type=str,
        choices=["ratio", "contrast", "both"],
        help="Method for computing forward and backward transfer. Defaults to ratio.",
    )

    # Method for normalization
    parser.add_argument(
        "-n",
        "--normalization-method",
        default="task",
        type=str,
        choices=["task", "run", "none"],
        help="Method for normalizing data. Defaults to task.",
    )

    # Method for smoothing
    parser.add_argument(
        "-g",
        "--smoothing-method",
        default="flat",
        type=str,
        choices=["flat", "hanning", "hamming", "bartlett", "blackman", "none"],
        help="Method for smoothing data, window type. Defaults to flat.",
    )

    # Flag for smoothing evaluation block data
    parser.add_argument(
        "-G",
        "--smooth-eval-data",
        dest="do_smooth_eval_data",
        default=False,
        action="store_true",
        help="Smooth evaluation block data. Defaults to false.",
    )

    # Window length for smoothing
    parser.add_argument(
        "-w",
        "--window-length",
        default=None,
        type=int,
        help="Window length for smoothing data. Defaults to None.",
    )

    # Flag for removing outliers
    parser.add_argument(
        "-x",
        "--clamp-outliers",
        action="store_true",
        help="Remove outliers in data for metrics by clamping to quantiles. Defaults \
                            to false.",
    )

    # Data range file for normalization
    parser.add_argument(
        "-d",
        "--data-range-file",
        default=None,
        type=str,
        help="JSON file containing task performance ranges for normalization. \
                            Defaults to None.",
    )

    # Mean and standard deviation for adding noise to log data
    parser.add_argument(
        "-N",
        "--noise",
        default=[0, 0],
        metavar=("MEAN", "STD"),
        nargs=2,
        type=float,
        help="Mean and standard deviation for Gaussian noise in log data. Defaults \
                            to [0, 0].",
    )

    # Output directory
    parser.add_argument(
        "-O",
        "--output-dir",
        default="results",
        type=str,
        help="Directory for output files. Defaults to results.",
    )

    # Output filename
    parser.add_argument(
        "-o",
        "--output",
        default=None,
        type=str,
        help="Specify output filename for plot and results. Defaults to None.",
    )

    # Flag for showing evaluation block lines
    parser.add_argument(
        "-e",
        "--show-eval-lines",
        dest="show_eval_lines",
        default=True,
        action="store_true",
        help="Show lines between evaluation blocks. Defaults to true.",
    )
    parser.add_argument(
        "--no-show-eval-lines",
        dest="show_eval_lines",
        action="store_false",
        help="Do not show lines between evaluation blocks",
    )

    # Flag for enabling/disabling plotting
    parser.add_argument(
        "-P",
        "--do-plot",
        dest="do_plot",
        default=True,
        action="store_true",
        help="Plot performance. Defaults to true.",
    )
    parser.add_argument(
        "--no-plot",
        dest="do_plot",
        action="store_false",
        help="Do not plot performance",
    )

    # Flag for enabling/disabling save
    parser.add_argument(
        "-S",
        "--do-save",
        dest="do_save",
        default=True,
        action="store_true",
        help="Save metrics outputs. Defaults to true.",
    )
    parser.add_argument(
        "--no-save",
        dest="do_save",
        action="store_false",
        help="Do not save metrics outputs",
    )

    # Settings file arguments
    parser.add_argument(
        "-c",
        "--load-settings",
        default=None,
        type=str,
        help="Load L2Metrics settings from JSON file. Defaults to None.",
    )
    parser.add_argument(
        "-C",
        "--do-save-settings",
        dest="do_save_settings",
        default=True,
        action="store_true",
        help="Save L2Metrics settings to JSON file. Defaults to true.",
    )
    parser.add_argument(
        "--no-save-settings",
        dest="do_save_settings",
        action="store_false",
        help="Do not save L2Metrics settings to JSON file",
    )

    return parser



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import json
import logging
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
from tqdm import tqdm

from l2metrics import util
from l2metrics.parser import init_parser
from l2metrics.report import MetricsReport

logging.captureWarnings(True)
logger = logging.getLogger("l2metrics.__main__")


def run() -> None:
    # Initialize parser
    parser = init_parser()

    # Parse arguments
    args = parser.parse_args()
    kwargs = vars(args)

    if args.load_settings:
        with open(args.load_settings, "r") as settings_file:
            kwargs.update(json.load(settings_file))

    kwargs["output_dir"] = Path(args.output_dir)

    # Create output directory if it doesn't exist
    if (args.do_save or args.do_save_settings) and args.ste_store_mode is None:
        args.output_dir.mkdir(parents=True, exist_ok=True)

    # Load data range data for normalization and standardize names to lowercase
    if args.data_range_file:
        with open(args.data_range_file) as data_range_file:
            data_range = json.load(data_range_file)
            data_range = {key.lower(): val for key, val in data_range.items()}
    else:
        data_range = None
    kwargs["data_range"] = data_range

    # Check for recursive flag
    if args.recursive:
        ll_metrics_df = pd.DataFrame()
        ll_metrics_dicts = []
        regime_metrics_df = pd.DataFrame()
        log_data_df = pd.DataFrame()
        task_colors = {}
        cc = util.color_cycler()

        # Iterate over all runs found in the directory
        dirs = [p for p in Path(args.log_dir).rglob("*") if p.is_dir()]
        for dir in tqdm(dirs, desc=Path(args.log_dir).name):
            # Check if current path is log directory for single run
            if all(
                x in [f.name for f in dir.glob("*.json")]
                for x in ["logger_info.json", "scenario_info.json"]
            ):
                if args.ste_store_mode:
                    # Store STE data
                    try:
                        util.store_ste_data(log_dir=dir, mode=args.ste_store_mode)
                    except ValueError as e:
                        logger.error(e)
                else:
                    # Compute and store the LL metrics
                    kwargs["log_dir"] = dir
                    report = MetricsReport(**kwargs)
                    report.calculate()
                    ll_metrics_df = ll_metrics_df.append(
                        report.ll_metrics_df, ignore_index=True
                    )
                    ll_metrics_dicts.append(report.ll_metrics_dict)
                    regime_metrics_df = regime_metrics_df.append(
                        report.regime_metrics_df, ignore_index=True
                    )
                    log_data_df = report._log_data
                    log_data_df["run_id"] = dir.name
                    log_data_df = log_data_df.append(log_data_df, ignore_index=True)

                    # Plot metrics
                    if args.do_plot:
                        # Update task color dictionary
                        for task_name, c in zip(
                            list(set(report._unique_tasks) - set(task_colors.keys())),
                            cc,
                        ):
                            task_colors[task_name] = c["color"]

                        # Generate plots
                        report.plot(
                            save=args.do_save,
                            show_eval_lines=args.show_eval_lines,
                            output_dir=str(args.output_dir),
                            task_colors=task_colors,
                        )
                        report.plot_ste_data(
                            save=args.do_save,
                            output_dir=str(args.output_dir),
                            task_colors=task_colors,
                        )
                        plt.close("all")

        # Assign base filename
        filename = args.output_dir / (args.output if args.output else "ll_metrics")

        # Save settings used to run calculate metrics
        if args.do_save_settings and args.ste_store_mode is None:
            with open(str(filename) + "_settings.json", "w") as settings_file:
                kwargs["log_dir"] = str(kwargs.get("log_dir", ""))
                kwargs["output_dir"] = str(kwargs.get("output_dir", ""))
                json.dump(kwargs, settings_file)

        # Save data
        if args.do_save and args.ste_store_mode is None:
            if not ll_metrics_df.empty:
                with open(str(filename) + ".tsv", "w", newline="\n") as metrics_file:
                    ll_metrics_df.set_index(["run_id"]).to_csv(metrics_file, sep="\t")
            if ll_metrics_dicts:
                with open(str(filename) + ".json", "w", newline="\n") as metrics_file:
                    json.dump(ll_metrics_dicts, metrics_file)
            if not regime_metrics_df.empty:
                with open(
                    str(filename) + "_regime.tsv", "w", newline="\n"
                ) as metrics_file:
                    regime_metrics_df.set_index(["run_id"]).to_csv(
                        metrics_file, sep="\t"
                    )
            if not log_data_df.empty:
                log_data_df.reset_index(drop=True).to_feather(
                    str(filename) + "_data.feather"
                )
    else:
        if args.ste_store_mode:
            # Store STE data
            util.store_ste_data(log_dir=Path(args.log_dir), mode=args.ste_store_mode)
        else:
            # Initialize metrics report
            report = MetricsReport(**kwargs)

            # Add noise to log data if mean or standard deviation is specified
            if args.noise[0] or args.noise[1]:
                report.add_noise(mean=args.noise[0], std=args.noise[1])

            # Calculate metrics in order of their addition to the metrics list.
            report.calculate()

            # Print table of metrics
            report.report()

            # Save metrics to file
            if args.do_save:
                report.save_metrics(output_dir=args.output_dir, filename=args.output)
                report.save_data(output_dir=args.output_dir, filename=args.output)

            # Plot metrics
            if args.do_plot:
                report.plot(
                    save=args.do_save,
                    show_eval_lines=args.show_eval_lines,
                    output_dir=str(args.output_dir),
                )
                report.plot_ste_data(save=args.do_save, output_dir=str(args.output_dir))
                plt.show()

            # Save settings used to run calculate metrics
            if args.do_save_settings:
                report.save_settings(output_dir=args.output_dir, filename=args.output)


if __name__ == "__main__":
    # Configure logger
    handler = logging.StreamHandler()
    formatter = logging.Formatter(
        fmt="%(asctime)s | %(levelname)-8s | %(name)s | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    handler.setFormatter(formatter)
    handler.setStream(tqdm)
    handler.terminator = ""

    logging.basicConfig(level=logging.INFO, handlers=[handler])

    try:
        run()
    except (KeyError, ValueError) as e:
        logger.exception(e)



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import numpy as np
import pandas as pd
from l2metrics._localutil import smooth
from l2metrics.normalizer import Normalizer


def test_smoothing():
    x = np.array((1, 2, 3, 4, 5, 6, 7, 8, 9, 10))
    assert np.allclose(
        smooth(x, window_len=3),
        (2, 1.66666667, 2, 3, 4, 5, 6, 7, 8, 9),
        rtol=1e-05,
        atol=1e-08,
    )


def test_clamp_outliers_quantiles():
    x = np.linspace(0, 100, 100)
    lower_bound, upper_bound = np.quantile(x, (0.1, 0.9))
    clamped_x = x.clip(lower_bound, upper_bound)
    assert (lower_bound, upper_bound) == (clamped_x.min(), clamped_x.max())


def test_clamp_outliers_range():
    x = np.linspace(0, 100, 100)
    lower_bound = 10
    upper_bound = 90
    clamped_x = x.clip(lower_bound, upper_bound)
    assert (lower_bound, upper_bound) == (clamped_x.min(), clamped_x.max())


def test_normalization_data_range():
    # Initialise data of lists
    data = {
        "task_name": ["t1", "t1", "t1", "t2", "t2", "t2"],
        "performance": [10, 15, 30, 40, 45, 50],
    }
    df = pd.DataFrame(data)

    normalizer = Normalizer(
        perf_measure="performance",
        data=df[["task_name", "performance"]].set_index("task_name"),
    )

    assert (normalizer.run_min, normalizer.run_max, normalizer.data_range) == (
        10,
        50,
        {"t1": {"min": 10, "max": 30}, "t2": {"min": 40, "max": 50}},
    )


def test_normalization():
    # Initialise data of lists
    data = {"task_name": ["t1", "t1", "t1"], "performance": [-100, 0, 100]}
    df = pd.DataFrame(data)

    normalizer = Normalizer(
        perf_measure="performance",
        data=df[["task_name", "performance"]].set_index("task_name"),
    )

    assert (normalizer.normalize(df)["performance"].to_numpy() == (1, 51, 101)).all()



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging
import os
import pickle
from collections import OrderedDict
from math import ceil
from pathlib import Path
from typing import List

import l2logger.util as l2l
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# Get default color cycler
color_cycler = plt.rcParams["axes.prop_cycle"]

logger = logging.getLogger(__name__)


def get_ste_data_names() -> list:
    """Gets the names of the stored STE data in $L2DATA/taskinfo/.

    Returns:
        list: The STE task names.
    """

    ste_files = list(Path(l2l.get_l2root_base_dirs("taskinfo")).glob("*.pickle"))

    if ste_files:
        return np.char.lower([f.stem for f in ste_files])
    else:
        return []


def load_ste_data(task_name: str) -> List[pd.DataFrame]:
    """Loads the STE data corresponding to the given task name.

    Args:
        task_name (str): The name of the STE data file.

    Returns:
        List[pd.DataFrame]: The STE data if found, else empty list.
    """

    # Variant-aware STE task names
    ste_task_variant_names = get_ste_data_names()

    # Variant-agnostic STE task names
    ste_task_base_names = set(
        [task_name.split("_")[0] for task_name in ste_task_variant_names]
    )

    if task_name in ste_task_variant_names:
        # Load variant-aware STE data
        ste_file_name = l2l.get_l2root_base_dirs("taskinfo", task_name + ".pickle")
        with open(ste_file_name, "rb") as ste_file:
            ste_data = pickle.load(ste_file)
            return ste_data
    elif task_name in ste_task_base_names:
        ste_data = []
        # Load variant-agnostic STE data
        for ste_variant_file in l2l.get_l2root_base_dirs("taskinfo").glob(
            task_name + "*.pickle"
        ):
            with open(ste_variant_file, "rb") as ste_file:
                ste_data.extend(pickle.load(ste_file))

        # Remove variant label from task names
        for idx, ste_data_df in enumerate(ste_data):
            ste_data[idx]["task_name"] = ste_data_df["task_name"].apply(
                lambda x: x.split("_")[0]
            )

        return ste_data
    else:
        return []


def store_ste_data(log_dir: Path, mode: str = "w") -> None:
    """Stores the STE data in the given log directory as a serialized DataFrame.

    Args:
        log_dir (Path): The log directory of the STE data.
        mode (str, optional): The mode for saving STE data. Defaults to 'w'.
            'w' - Write - Opens a file for writing, overwrites data if the file exists.
            'a' - Append - Opens a file for writing, appends data if the file exists.

    Raises:
        ValueError: If scenario does not only contain one task for training.
    """

    # Load data from ste logs
    ste_data_df = l2l.read_log_data(log_dir)

    # Get metric fields
    logger_info = l2l.read_logger_info(log_dir)

    # Validate data format
    l2l.validate_log(ste_data_df, logger_info["metrics_columns"])

    # Filter data by completed experiences
    ste_data_df = ste_data_df[ste_data_df["exp_status"] == "complete"]

    # Fill in regime number and sort
    ste_data_df = l2l.fill_regime_num(ste_data_df)
    ste_data_df = ste_data_df.sort_values(by=["regime_num", "exp_num"])

    # Get training task name
    task_name = list(
        ste_data_df[ste_data_df["block_type"] == "train"].task_name.unique()
    )

    # Check for number of tasks in scenario
    if len(task_name) != 1:
        raise ValueError(
            f"Expected 1 trained task in {log_dir.name} but found {len(task_name)}"
        )

    # Add STE dataframe to list
    ste_data = [ste_data_df]

    # Create task info directory if it doesn't exist
    task_info_dir = l2l.get_l2root_base_dirs("taskinfo")
    if not task_info_dir.exists():
        task_info_dir.mkdir(parents=True, exist_ok=True)

    # Get base directory to store ste data
    filename = task_info_dir / (task_name[0] + ".pickle")

    # Store ste data in task info directory
    if mode == "a":
        # Load existing STE data and append
        if filename.exists():
            with open(filename, "rb") as ste_file:
                stored_ste_data = pickle.load(ste_file)
                ste_data.extend(stored_ste_data)

    # Write/Overwrite STE data to file
    with open(filename, "wb") as ste_file:
        pickle.dump(ste_data, ste_file)

    logger.info(f"Stored STE data for {task_name[0]} in {log_dir.name}")


def plot_blocks(
    dataframe: pd.DataFrame,
    reward: str,
    unique_tasks: list,
    task_colors: dict = {},
    input_title: str = "",
    output_dir: str = "",
    do_save_fig: bool = False,
    plot_filename: str = "block_plot",
):
    """Plot learning performance curves and evaluation blocks as separate plots.

    Args:
        dataframe (pd.DataFrame): The performance data to plot.
        reward (str): The column name of the metric to plot.
        unique_tasks (list): List of unique tasks in scenario.
        task_colors (dict): Dict of task names and colors for plotting. Defaults to {}.
        input_title (str, optional): The plot title. Defaults to ''.
        output_dir (str, optional): Output directory of results. Defaults to ''.
        do_save_fig (bool, optional): Flag for enabling saving figure. Defaults to False.
        plot_filename (str, optional): The filename to use for saving. Defaults to 'block_plot'.
    """

    reward_col = reward + "_raw"

    if reward + "_smoothed" in dataframe.columns:
        reward_col_smooth = reward + "_smoothed"
    else:
        reward_col_smooth = None

    # Use sleep evaluation blocks if they exist and filter out wake evaluation
    if "sleep" in dataframe["block_subtype"].to_numpy():
        dataframe = dataframe[
            ~(
                dataframe.block_type.isin(["test"])
                & dataframe.block_subtype.isin(["wake"])
            )
        ]

    df_test = dataframe[dataframe.block_type == "test"]
    df_train = dataframe[dataframe.block_type == "train"]

    fig, axes = plt.subplots(
        len(unique_tasks) + 1, 1, figsize=(12, 12), sharex=True, constrained_layout=True
    )
    ax0 = axes[0]
    ax0.set_ylabel(reward_col + " (LX)")
    ax0.grid()

    task_idx = 1
    xv_max = None  # Workaround for inconsistent # of samples

    # Assign colors for each task
    if not task_colors:
        task_colors = {
            task: c["color"] for c, task in zip(color_cycler(), unique_tasks)
        }

    # Plot training and test data
    for task_name in unique_tasks:
        x = df_train[df_train["task_name"] == task_name].exp_num
        y = df_train[df_train["task_name"] == task_name][reward_col]
        ax0.plot(x, y, ".", label=task_name, color=task_colors[task_name], markersize=4)

        ax = axes[task_idx]
        x = df_test[df_test["task_name"] == task_name].exp_num
        y = df_test[df_test["task_name"] == task_name][reward_col]
        ax.plot(x, y, ".", label=task_name, color=task_colors[task_name], markersize=4)
        task_ex_block_data = df_test[df_test["task_name"] == task_name].groupby(
            "block_num"
        )
        xv = task_ex_block_data.exp_num.median()
        if xv_max is None or len(xv) > len(xv_max):
            xv_max = xv
        ex_median = task_ex_block_data[reward_col].median()
        ex_iqr_lower = task_ex_block_data[reward_col].quantile(0.25)
        ex_iqr_upper = task_ex_block_data[reward_col].quantile(0.75)
        ax.fill_between(
            xv, ex_iqr_lower, ex_iqr_upper, color=task_colors[task_name], alpha=0.5
        )
        ax.plot(xv, ex_median, color=task_colors[task_name])
        ax.set_ylabel(task_name + " (EX)")
        ax.grid()
        task_idx += 1

    # Plot smoothed training data
    if reward_col_smooth is not None:
        for _, group in df_train.groupby("block_num"):
            ax0.plot(group.exp_num, group[reward_col_smooth], "k")

    # Set plot title
    if Path(input_title).parent != Path("."):
        _, input_title = os.path.split(input_title)

    ax0.set_title(input_title)

    # Enable plot legend
    ax0.legend()

    # Set y-axis limits
    if not df_test.empty:
        plt.setp(
            fig.axes,
            ylim=(np.nanmin(df_test[reward_col]), np.nanmax(df_test[reward_col])),
        )

    if do_save_fig:
        logger.info(f'Saving block plot with name: {plot_filename.replace(" ", "_")}')
        fig.savefig(Path(output_dir) / (plot_filename.replace(" ", "_") + ".png"))


def plot_performance(
    dataframe: pd.DataFrame,
    block_info: pd.DataFrame,
    unique_tasks: list,
    task_colors: dict = {},
    x_axis_col: str = "exp_num",
    y_axis_col: str = "reward",
    input_title: str = "",
    input_xlabel: str = "Experiences",
    input_ylabel: str = "Performance",
    show_eval_lines: bool = True,
    show_block_boundary: bool = False,
    shade_test_blocks: bool = True,
    output_dir: str = "",
    do_save_fig: bool = False,
    plot_filename: str = "performance_plot",
) -> None:
    """Plots the performance curves for the given DataFrame.

    Args:
        dataframe (pd.DataFrame): The performance data to plot.
        block_info (pd.DataFrame): The block info of the DataFrame.
        unique_tasks (list): List of unique tasks in scenario.
        task_colors (dict): Dict of task names and colors for plotting. Defaults to {}.
        x_axis_col (str, optional): The column name of the x-axis data. Defaults to 'exp_num'.
        y_axis_col (str, optional): The column name of the metric to plot. Defaults to 'reward'.
        input_title (str, optional): The plot title. Defaults to ''.
        input_xlabel (str, optional): The x-axis label. Defaults to 'Experiences'.
        input_ylabel (str, optional): The y-axis label. Defaults to 'Performance'.
        show_eval_lines (bool, optional): Flag for enabling lines between evaluation blocks to show
            changing slope of evaluation performance. Defaults to True.
        show_block_boundary (bool, optional): Flag for enabling block boundaries. Defaults to True.
        shade_test_blocks (bool, optional): Flag for enabling block shading. Defaults to True.
        output_dir (str, optional): Output directory of results. Defaults to ''.
        do_save_fig (bool, optional): Flag for enabling saving figure. Defaults to False.
        plot_filename (str, optional): The filename to use for saving. Defaults to 'performance_plot'.
    """

    # Initialize figure
    fig = plt.figure(figsize=(12, 6))
    ax = fig.add_subplot(111)

    # Use sleep evaluation blocks if they exist and filter out wake evaluation
    if "sleep" in block_info["block_subtype"].to_numpy():
        block_info = block_info[
            ~(
                block_info.block_type.isin(["test"])
                & block_info.block_subtype.isin(["wake"])
            )
        ]
        dataframe = dataframe[
            ~(
                dataframe.block_type.isin(["test"])
                & dataframe.block_subtype.isin(["wake"])
            )
        ]

    # Assign colors for each task
    if not task_colors:
        task_colors = {
            task: c["color"] for c, task in zip(color_cycler(), unique_tasks)
        }

    # Loop through tasks and plot their performance curves
    for task_name in unique_tasks:
        if show_eval_lines:
            eval_x_data = []
            eval_y_data = []
            (eval_line,) = ax.plot(
                eval_x_data,
                eval_y_data,
                color=task_colors[task_name],
                linestyle="--",
                alpha=0.2,
            )

        for _, row in block_info[block_info["task_name"] == task_name].iterrows():
            regime_num = row["regime_num"]
            block_type = row["block_type"]

            # Get data for current regime
            x = dataframe.loc[
                dataframe["regime_num"] == regime_num, x_axis_col
            ].to_numpy()
            y = dataframe.loc[
                dataframe["regime_num"] == regime_num, y_axis_col
            ].to_numpy()

            if show_block_boundary:
                ax.axes.axvline(x[0], linewidth=1, linestyle=":")

            if shade_test_blocks and block_type == "test":
                ax.axvspan(x[0], x[-1] + 1, alpha=0.1, facecolor="black")

            if block_type == "test":
                if show_eval_lines:
                    x = list(range(x[0], x[0] + len(y)))
                    eval_x_data.extend(x)
                    eval_y_data.extend(np.nanmean(y) * np.ones(len(x)))
                    eval_line.set_data(eval_x_data, eval_y_data)
                    plt.draw()

            # Match smoothed x and y length if data had NaNs
            if len(x) != len(y):
                x = list(range(x[0], x[0] + len(y)))

            ax.scatter(
                x, y, color=task_colors[task_name], marker="*", s=8, label=task_name
            )

    handles, labels = plt.gca().get_legend_handles_labels()
    by_label = OrderedDict(zip(labels, handles))
    ax.legend(by_label.values(), by_label.keys())

    if Path(input_title).parent != Path("."):
        _, input_title = os.path.split(input_title)

    # Want the saved figured to have a grid so do this before saving
    ax.set(xlabel=input_xlabel, ylabel=input_ylabel, title=input_title)
    ax.grid()
    fig.tight_layout()

    if do_save_fig:
        logger.info(
            f'Saving performance plot with name: {plot_filename.replace(" ", "_")}'
        )
        fig.savefig(Path(output_dir) / (plot_filename.replace(" ", "_") + ".png"))


def plot_ste_data(
    dataframe: pd.DataFrame,
    ste_data: dict,
    block_info: pd.DataFrame,
    unique_tasks: list,
    task_colors: dict = {},
    perf_measure: str = "reward",
    ste_averaging_method: str = "metrics",
    input_title: str = "",
    input_xlabel: str = "Experiences",
    input_ylabel: str = "Performance",
    output_dir: str = "",
    do_save: bool = False,
    plot_filename: str = "ste_plot",
) -> None:
    """Plots the relative performance of tasks compared to Single-Task Experts.

    Args:
        dataframe (pd.DataFrame): The performance data to plot.
        ste_data (dict): STE data.
        block_info (pd.DataFrame): The block info of the DataFrame.
        unique_tasks (list): List of unique tasks in scenario.
        task_colors (dict): Dict of task names and colors for plotting. Defaults to {}.
        perf_measure (str, optional): The column name of the metric to plot. Defaults to 'reward'.
        ste_averaging_method (str, optional): Method for handling STE metric averaging. Defaults to 'metrics'.
        input_title (str, optional): Plot title. Defaults to ''.
        input_xlabel (str, optional): The x-axis label. Defaults to 'Experiences'.
        input_ylabel (str, optional): The y-axis label. Defaults to 'Performance'.
        output_dir (str, optional): Output directory of results. Defaults to ''.
        do_save (bool, optional): Flag for enabling saving figure. Defaults to False.
        plot_filename (str, optional): The filename to use for saving. Defaults to 'ste_plot'.
    """

    # Initialize figure
    fig = plt.figure(figsize=(12, 6))
    fig.suptitle(input_title)

    # Calculate subplot dimensions
    cols = min(len(unique_tasks), 2)
    rows = ceil(len(unique_tasks) / cols)

    # Initialize axis limits
    x_limit = 0
    y_limit = (np.nan, np.nan)

    # Assign colors for each task
    if not task_colors:
        task_colors = {
            task: c["color"] for c, task in zip(color_cycler(), unique_tasks)
        }

    for index, task_name in enumerate(unique_tasks):
        # Get block info for task during training
        task_blocks = block_info[
            (block_info["task_name"] == task_name)
            & (block_info["block_type"] == "train")
            & (block_info["block_subtype"] == "wake")
        ]

        # Get concatenated data for task
        task_data = dataframe[
            dataframe["regime_num"].isin(task_blocks["regime_num"])
        ].reset_index(drop=True)

        if len(task_data):
            # Load STE data
            if ste_data.get(task_name):
                # Create subplot
                ax = fig.add_subplot(rows, cols, index + 1)

                plt.scatter(
                    [],
                    [],
                    label=task_name,
                    color=task_colors[task_name],
                    marker="*",
                    s=8,
                )
                plt.scatter([], [], label="STE", color="orange", marker="*", s=8)

                # Plot LL data
                y_ll = task_data[perf_measure].to_numpy()
                x_ll = list(range(0, len(y_ll)))
                ax.scatter(
                    x_ll, y_ll, color=task_colors[task_name], marker="*", s=8, zorder=3
                )

                # Get STE data
                y_ste = [
                    ste_data_df[ste_data_df["block_type"] == "train"][
                        perf_measure
                    ].to_numpy()
                    for ste_data_df in ste_data.get(task_name)
                ]

                if ste_averaging_method == "time":
                    # Average all the STE data together after truncating to same length
                    y_ste = np.array([x[: min(map(len, y_ste))] for x in y_ste]).mean(0)
                    x_ste = list(range(0, len(y_ste)))
                    ax.scatter(x_ste, y_ste, color="orange", marker="*", s=8)

                    x_limit = max(x_limit, len(y_ste), len(y_ll))
                    y_limit = (
                        np.nanmin([y_limit[0], np.nanmin(y_ste), np.nanmin(y_ll)]),
                        np.nanmax([y_limit[1], np.nanmax(y_ste), np.nanmax(y_ll)]),
                    )
                else:
                    # Plot runs of STE data
                    for y in y_ste:
                        x = list(range(0, len(y)))
                        ax.scatter(x, y, color="orange", marker="*", s=8)
                        x_limit = max(x_limit, len(y), len(y_ll))
                        y_limit = (
                            np.nanmin([y_limit[0], np.nanmin(y), np.nanmin(y_ll)]),
                            np.nanmax([y_limit[1], np.nanmax(y), np.nanmax(y_ll)]),
                        )

                # Draw line at block boundaries of task data
                for x_val in task_data[task_data.regime_num.diff() != 0].index.tolist():
                    ax.axes.axvline(x=x_val, color="black", linestyle="--")

                ax.set(xlabel=input_xlabel, ylabel=input_ylabel)
                ax.grid()
                plt.legend()
            else:
                logger.warning(f"STE data for task cannot be found: {task_name}")
        else:
            logger.warning(
                f"Scenario does not contain training data for task: {task_name}"
            )

    fig.subplots_adjust(wspace=0.3, hspace=0.4)

    plt.setp(fig.axes, xlim=(0, x_limit), ylim=y_limit)
    fig.tight_layout()

    if do_save:
        logger.info(f'Saving STE plot with name: {plot_filename.replace(" ", "_")}')
        fig.savefig(Path(output_dir) / (plot_filename.replace(" ", "_") + ".png"))



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging

import numpy as np
import pandas as pd

from ._localutil import fill_metrics_df
from .core import Metric

logger = logging.getLogger(__name__)


class PerformanceMaintenance(Metric):
    name = "Performance Maintenance"
    capability = "adapting_to_new_tasks"
    requires = {"syllabus_type": "agent"}
    description = (
        "Calculates the average difference between the most recent"
        "terminal learning performance of a task and each evaluation performance"
    )

    def __init__(self, perf_measure: str, method: str = "mrlep") -> None:
        super().__init__()
        self.perf_measure = perf_measure

        # Check for valid method
        if method not in ["mrtlp", "mrlep", "both"]:
            raise KeyError(f"Invalid performance maintenance method: {method}")
        else:
            self.do_mrtlp = method in ["mrtlp", "both"]
            self.do_mrlep = method in ["mrlep", "both"]

    def validate(self, block_info: pd.DataFrame) -> None:
        # Initialize variables for checking block type format
        last_block_num = -1
        last_block_type = ""

        # Ensure alternating block types
        for _, regime in block_info.iterrows():
            if regime["block_num"] != last_block_num:
                last_block_num = regime["block_num"]

                if regime["block_type"] == "test":
                    if last_block_type == "test":
                        logger.warning("Block types are not alternating")
                        break
                    last_block_type = "test"
                elif regime["block_type"] == "train":
                    if last_block_type == "train":
                        logger.warning("Block types are not alternating")
                        break
                    last_block_type = "train"

    def calculate(
        self,
        dataframe: pd.DataFrame,
        block_info: pd.DataFrame,
        metrics_df: pd.DataFrame,
    ) -> pd.DataFrame:
        # Validate block structure
        self.validate(block_info)

        # Use sleep evaluation blocks if they exist
        if "sleep" in block_info["block_subtype"].to_numpy():
            block_info_df = block_info[
                ~(
                    block_info.block_type.isin(["test"])
                    & block_info.block_subtype.isin(["wake"])
                )
            ]
        else:
            block_info_df = block_info.copy()

        # Initialize metric columns
        maintenance_values_mrtlp = {}
        maintenance_values_mrlep = {}
        pm_values_mrtlp = {}
        pm_values_mrlep = {}

        # Get unique tasks in scenario
        unique_tasks = block_info_df.task_name.unique()

        # Iterate over tasks
        for task in unique_tasks:
            # Get training and test regimes
            training_regs = block_info_df[
                (block_info_df["task_name"] == task)
                & (block_info_df["block_type"] == "train")
            ]["regime_num"].to_numpy()

            test_regs = block_info_df[
                (block_info_df["task_name"] == task)
                & (block_info_df["block_type"] == "test")
            ]["regime_num"].to_numpy()

            # Get reference test regimes
            ref_test_regs = np.array(
                [
                    test_regs[np.argmax(test_regs > x)]
                    for x in training_regs
                    if np.any(test_regs > x)
                ]
            )

            # Iterate over test regimes
            for test_regime in test_regs:
                # Check that current test block occurred after last reference test
                if (
                    np.any(test_regime > ref_test_regs)
                    and test_regime not in ref_test_regs
                ):
                    # Get performance of current test regime
                    test_perf = metrics_df["term_perf"][test_regime]

                    if self.do_mrtlp:
                        ref_regime = training_regs[test_regime > training_regs][-1]
                        mrtlp = metrics_df["term_perf"][ref_regime]
                        maintenance_values_mrtlp[test_regime] = test_perf - mrtlp
                    if self.do_mrlep:
                        ref_regime = ref_test_regs[test_regime > ref_test_regs][-1]
                        mrlep = metrics_df["term_perf"][ref_regime]
                        maintenance_values_mrlep[test_regime] = test_perf - mrlep

        if self.do_mrtlp and maintenance_values_mrtlp:
            # Fill metrics dataframe with most recent terminal learning performance differences
            metrics_df = fill_metrics_df(
                maintenance_values_mrtlp, "maintenance_val_mrtlp", metrics_df
            )

            # Iterate over task performance differences for performance maintenance
            for task in unique_tasks:
                # Get the task maintenance values
                m = metrics_df[metrics_df["task_name"] == task][
                    "maintenance_val_mrtlp"
                ].to_numpy(dtype=float)

                # Remove NaNs
                m = m[~np.isnan(m)]

                # Calculate performance maintenance value
                if m.size:
                    pm_values_mrtlp[
                        block_info_df.index[block_info_df["task_name"] == task][-1]
                    ] = np.mean(m)

            metrics_df = fill_metrics_df(
                pm_values_mrtlp, "perf_maintenance_mrtlp", metrics_df
            )

        if self.do_mrlep and maintenance_values_mrlep:
            # Fill metrics dataframe with most recent terminal learning performance differences
            metrics_df = fill_metrics_df(
                maintenance_values_mrlep, "maintenance_val_mrlep", metrics_df
            )

            # Iterate over task performance differences for performance maintenance
            for task in unique_tasks:
                # Get the task maintenance values
                m = metrics_df[metrics_df["task_name"] == task][
                    "maintenance_val_mrlep"
                ].to_numpy(dtype=float)

                # Remove NaNs
                m = m[~np.isnan(m)]

                # Calculate performance maintenance value
                if m.size:
                    pm_values_mrlep[
                        block_info_df.index[block_info_df["task_name"] == task][-1]
                    ] = np.mean(m)

            metrics_df = fill_metrics_df(
                pm_values_mrlep, "perf_maintenance_mrlep", metrics_df
            )

        return metrics_df



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import abc

import pandas as pd


class Metric(abc.ABC):
    """
    A single metric
    """

    @property
    def capability(self):
        """
        A string (one of the core capabilities that the metric calculates):
            continual_learning
            adapt_to_new_tasks
            goal_driven_perception
            selective_plasticity
            safety_and_monitoring
        """
        pass

    @property
    def name(self):
        """
        A short label that uniquely identifies this metric.
        """
        return ""

    @property
    def description(self):
        """
        A more detailed description for this metric
        """
        return ""

    @property
    def requires(self):
        """
        A dictionary of requirements for this metric. Keys
          syllabus_type: one of 'type1", "type2", etc.
        """
        return {}

    @abc.abstractmethod
    def calculate(
        self,
        dataframe: pd.DataFrame,
        block_info: pd.DataFrame,
        metrics_df: pd.DataFrame,
    ) -> pd.DataFrame:
        """Calculate metric

        Args:
            dataframe (pd.DataFrame): Dataframe containing log data
            block_info (pd.DataFrame): High-level block summary of log data
            metrics_df (pd.DataFrame): Incremental Dataframe with columns corresponding to
                calculated metrics along with some of the block_info information.

        Returns:
            pd.DataFrame: Updated metrics dataframe.
        """
        return None



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging
from itertools import permutations

import numpy as np
import pandas as pd

from ._localutil import fill_metrics_df
from .core import Metric

logger = logging.getLogger(__name__)


class Transfer(Metric):
    name = "Transfer"
    capability = "adapt_to_new_tasks"
    requires = {"syllabus_type": "agent"}
    description = "Calculates the forward and backward transfer for valid task pairs"

    def __init__(self, perf_measure: str, method: str = "contrast") -> None:
        super().__init__()
        self.perf_measure = perf_measure

        # Check for valid transfer method
        if method not in ["ratio", "contrast", "both"]:
            raise KeyError(f"Invalid transfer method: {method}")
        else:
            self.do_ratio = method in ["ratio", "both"]
            self.do_contrast = method in ["contrast", "both"]

    def validate(self, block_info: pd.DataFrame) -> None:
        # Initialize variables for checking block type format
        last_block_num = -1
        last_block_type = ""

        # Ensure alternating block types
        for _, regime in block_info.iterrows():
            if regime["block_num"] != last_block_num:
                last_block_num = regime["block_num"]

                if regime["block_type"] == "test":
                    if last_block_type == "test":
                        logger.warning("Block types are not alternating")
                        break
                    last_block_type = "test"
                elif regime["block_type"] == "train":
                    if last_block_type == "train":
                        logger.warning("Block types are not alternating")
                        break
                    last_block_type = "train"

    def calculate(
        self,
        dataframe: pd.DataFrame,
        block_info: pd.DataFrame,
        metrics_df: pd.DataFrame,
    ) -> pd.DataFrame:
        try:
            # Validate data
            self.validate(block_info)

            # Use sleep evaluation blocks if they exist
            if "sleep" in block_info["block_subtype"].to_numpy():
                block_info_df = block_info[
                    ~(
                        block_info.block_type.isin(["test"])
                        & block_info.block_subtype.isin(["wake"])
                    )
                ]
            else:
                block_info_df = block_info.copy()

            # Initialize metric dictionaries
            forward_transfer = {"ratio": {}, "contrast": {}}
            backward_transfer = {"ratio": {}, "contrast": {}}

            # Find eligible tasks for forward transfer
            unique_tasks = block_info_df.task_name.unique()

            # Determine valid transfer pairs
            for task, other_task in permutations(unique_tasks, 2):
                # Get testing and training indices for task pair
                training_regs = block_info_df[
                    (block_info_df["task_name"] == task)
                    & (block_info_df["block_type"] == "train")
                ]["regime_num"].to_numpy()

                other_blocks = block_info_df[block_info_df["task_name"] == other_task]
                other_test_regs = other_blocks[other_blocks["block_type"] == "test"][
                    "regime_num"
                ].to_numpy()
                other_training_regs = other_blocks[
                    other_blocks["block_type"] == "train"
                ]["regime_num"].to_numpy()

                # FT - Must have tested task 2 before training task 1 then tested task 2 again
                if len(training_regs):
                    # Get valid training regimes for first task forward transfer
                    valid_ft_training_regs = (
                        training_regs[training_regs < np.min(other_training_regs)]
                        if len(other_training_regs)
                        else training_regs
                    )

                    # Get valid training regimes for first task backward transfer
                    valid_bt_training_regs = (
                        training_regs[training_regs > np.min(other_training_regs)]
                        if len(other_training_regs)
                        else []
                    )

                    # Calculate forward transfer
                    for training_regime in valid_ft_training_regs:
                        for test_regime_1, test_regime_2 in zip(
                            other_test_regs, other_test_regs[1:]
                        ):
                            if test_regime_1 < training_regime < test_regime_2:
                                tp_1 = metrics_df[
                                    metrics_df["regime_num"] == test_regime_1
                                ]["term_perf"].to_numpy()[0]
                                tp_2 = metrics_df[
                                    metrics_df["regime_num"] == test_regime_2
                                ]["term_perf"].to_numpy()[0]

                                if self.do_ratio:
                                    if tp_1 != 0:
                                        forward_transfer["ratio"][test_regime_2] = {
                                            task: tp_2 / tp_1
                                        }
                                if self.do_contrast:
                                    if (tp_1 + tp_2) != 0:
                                        forward_transfer["contrast"][test_regime_2] = {
                                            task: (tp_2 - tp_1) / (tp_1 + tp_2)
                                        }

                    # Calculate backward transfer
                    for training_regime in valid_bt_training_regs:
                        for test_regime_1, test_regime_2 in zip(
                            other_test_regs, other_test_regs[1:]
                        ):
                            if test_regime_1 < training_regime < test_regime_2:
                                tp_1 = metrics_df[
                                    (metrics_df["regime_num"] == test_regime_1)
                                ]["term_perf"].to_numpy()[0]
                                tp_2 = metrics_df[
                                    (metrics_df["regime_num"] == test_regime_2)
                                ]["term_perf"].to_numpy()[0]

                                if self.do_ratio:
                                    if tp_1 != 0:
                                        backward_transfer["ratio"][test_regime_2] = {
                                            task: tp_2 / tp_1
                                        }
                                if self.do_contrast:
                                    if (tp_1 + tp_2) != 0:
                                        backward_transfer["contrast"][test_regime_2] = {
                                            task: (tp_2 - tp_1) / (tp_1 + tp_2)
                                        }

            if not (forward_transfer["ratio"] or forward_transfer["contrast"]):
                logger.warning("Cannot compute Forward Transfer - No valid task pairs")
            else:
                if self.do_ratio:
                    metrics_df = fill_metrics_df(
                        forward_transfer["ratio"], "forward_transfer_ratio", metrics_df
                    )
                if self.do_contrast:
                    metrics_df = fill_metrics_df(
                        forward_transfer["contrast"],
                        "forward_transfer_contrast",
                        metrics_df,
                    )

            if not (backward_transfer["ratio"] or backward_transfer["contrast"]):
                logger.warning("Cannot compute Backward Transfer - No valid task pairs")
            else:
                if self.do_ratio:
                    metrics_df = fill_metrics_df(
                        backward_transfer["ratio"],
                        "backward_transfer_ratio",
                        metrics_df,
                    )
                if self.do_contrast:
                    metrics_df = fill_metrics_df(
                        backward_transfer["contrast"],
                        "backward_transfer_contrast",
                        metrics_df,
                    )

            return metrics_df
        except ValueError as e:
            logger.warning(f"Cannot compute {self.name} - {e}")
            return metrics_df



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging

import numpy as np
import pandas as pd

from ._localutil import fill_metrics_df
from .core import Metric

logger = logging.getLogger(__name__)


class STERelativePerf(Metric):
    name = "Performance relative to STE"
    capability = "adapt_to_new_tasks"
    requires = {"syllabus_type": "agent"}
    description = "Calculates the performance of each task relative to it's corresponding single-task expert"

    def __init__(
        self, perf_measure: str, ste_data: dict, ste_averaging_method: str = "metrics"
    ) -> None:

        super().__init__()
        self.perf_measure = perf_measure
        self.ste_data = ste_data
        if ste_averaging_method not in ["time", "metrics"]:
            raise KeyError(f"Invalid STE averaging method: {ste_averaging_method}")
        else:
            self.ste_averaging_method = ste_averaging_method

    def validate(self, block_info: pd.DataFrame) -> None:
        # Check if there is STE data for each task in the scenario
        self.unique_tasks = block_info.loc[:, "task_name"].unique()
        ste_names = tuple(self.ste_data.keys())

        # Raise value error if none of the tasks have STE data
        if ~np.any(np.isin(self.unique_tasks, ste_names)):
            raise ValueError("No STE data available for any task")

        # Make sure STE baselines are available for all tasks, else log warning
        if ~np.all(np.isin(self.unique_tasks, ste_names)):
            logger.warning("STE data not available for all tasks")

    def calculate(
        self,
        dataframe: pd.DataFrame,
        block_info: pd.DataFrame,
        metrics_df: pd.DataFrame,
    ) -> pd.DataFrame:
        try:
            # Validate the STE
            self.validate(block_info)

            # Initialize metric dictionaries
            ste_rel_perf = {}

            for task in self.unique_tasks:
                # Get block info for task during training
                task_blocks = block_info[
                    (block_info["task_name"] == task)
                    & (block_info["block_type"] == "train")
                    & (block_info["block_subtype"] == "wake")
                ]

                # Get concatenated data for task
                task_data = dataframe[
                    dataframe.regime_num.isin(task_blocks["regime_num"])
                ]

                if len(task_data):
                    # Get STE data
                    ste_data = self.ste_data.get(task)

                    if ste_data:
                        if self.ste_averaging_method == "time":
                            # Average all the STE data together after truncating to same length
                            x_ste = [
                                ste_data_df[ste_data_df["block_type"] == "train"][
                                    self.perf_measure
                                ].to_numpy()
                                for ste_data_df in ste_data
                            ]
                            min_ste_exp = min(map(len, x_ste))
                            x_ste = np.array([x[:min_ste_exp] for x in x_ste]).mean(0)

                            # Compute relative performance
                            min_exp = min(task_data.shape[0], len(x_ste))
                            task_perf = np.nansum(
                                task_data.head(min_exp)[self.perf_measure].to_numpy()
                            )
                            ste_perf = np.nansum(x_ste[:min_exp])
                            rel_perf = task_perf / ste_perf
                            ste_rel_perf[task_data["regime_num"].iloc[-1]] = [rel_perf]
                        elif self.ste_averaging_method == "metrics":
                            rel_perf_vals = []

                            for ste_data_df in ste_data:
                                ste_data_df = ste_data_df[
                                    ste_data_df["block_type"] == "train"
                                ]

                                # Compute relative performance
                                min_exp = np.min(
                                    [task_data.shape[0], ste_data_df.shape[0]]
                                )
                                task_perf = np.nansum(
                                    task_data.head(min_exp)[
                                        self.perf_measure
                                    ].to_numpy()
                                )
                                ste_perf = np.nansum(
                                    ste_data_df.head(min_exp)[
                                        self.perf_measure
                                    ].to_numpy()
                                )
                                rel_perf_vals.append(task_perf / ste_perf)

                            ste_rel_perf[
                                task_data["regime_num"].iloc[-1]
                            ] = rel_perf_vals
                    else:
                        logger.warning(
                            f"Cannot compute {self.name} for task {task} - No STE data available"
                        )

            return fill_metrics_df(ste_rel_perf, "ste_rel_perf", metrics_df)
        except ValueError as e:
            logger.warning(f"Cannot compute {self.name} - {e}")
            return metrics_df



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import json
import logging
from collections import defaultdict
from datetime import datetime as dt
from pathlib import Path
from typing import List, Tuple, Union

import l2logger.util as l2l
import numpy as np
import pandas as pd
from tabulate import tabulate

from ._localutil import smooth
from .block_saturation import BlockSaturation
from .core import Metric
from .normalizer import Normalizer
from .performance_maintenance import PerformanceMaintenance
from .performance_recovery import PerformanceRecovery
from .recovery_time import RecoveryTime
from .sample_efficiency import SampleEfficiency
from .ste_relative_performance import STERelativePerf
from .terminal_performance import TerminalPerformance
from .transfer import Transfer
from .util import load_ste_data, plot_blocks, plot_performance, plot_ste_data

logger = logging.getLogger(__name__)


class MetricsReport:
    """
    Aggregates a list of metrics for an agent learner.
    """

    def __init__(self, **kwargs) -> None:
        # Defines log_dir and initializes the metrics list
        self._metrics = []
        self.ll_metrics_dict = {}

        self.log_dir = Path(kwargs.get("log_dir", ""))
        self.perf_measure = kwargs.get("perf_measure", "reward")
        self.variant_mode = kwargs.get("variant_mode", "aware")
        self.ste_averaging_method = kwargs.get("ste_averaging_method", "metrics")
        self.aggregation_method = kwargs.get("aggregation_method", "mean")
        self.maintenance_method = kwargs.get("maintenance_method", "mrlep")
        self.transfer_method = kwargs.get("transfer_method", "ratio")
        self.normalization_method = kwargs.get("normalization_method", "task")
        self.smoothing_method = kwargs.get("smoothing_method", "flat")
        self.do_smooth_eval_data = kwargs.get("do_smooth_eval_data", False)
        self.window_length = kwargs.get("window_length", None)
        self.clamp_outliers = kwargs.get("clamp_outliers", False)
        self.data_range = kwargs.get("data_range", None)

        # Modify data range based on variant mode
        if self.variant_mode == "agnostic" and self.data_range is not None:
            temp_data_range = {}
            for task_name in set(
                [variant_name.split("_")[0] for variant_name in self.data_range.keys()]
            ):
                task_ranges = [
                    data_range
                    for variant_name, data_range in self.data_range.items()
                    if task_name in variant_name
                ]
                temp_data_range[task_name] = {
                    "min": np.min([d["min"] for d in task_ranges]),
                    "max": np.max([d["max"] for d in task_ranges]),
                }
            self.data_range = temp_data_range

        # Initialize list of LL metrics
        self.task_metrics = ["perf_recovery"]
        if self.maintenance_method in ["mrlep", "both"]:
            self.task_metrics.extend(["perf_maintenance_mrlep"])
        if self.maintenance_method in ["mrtlp", "both"]:
            self.task_metrics.extend(["perf_maintenance_mrtlp"])
        if self.transfer_method in ["ratio", "both"]:
            self.task_metrics.extend(
                ["forward_transfer_ratio", "backward_transfer_ratio"]
            )
        if self.transfer_method in ["contrast", "both"]:
            self.task_metrics.extend(
                ["forward_transfer_contrast", "backward_transfer_contrast"]
            )
        self.task_metrics.extend(["ste_rel_perf", "sample_efficiency"])

        # Get metric fields
        self.logger_info = l2l.read_logger_info(self.log_dir)

        # Do a check to make sure the performance measure exists in logger info
        if self.perf_measure not in self.logger_info["metrics_columns"]:
            raise KeyError(
                f"Performance measure ({self.perf_measure}) not found in valid metrics columns: "
                f"{self.logger_info['metrics_columns']}"
            )

        # Gets all data from the relevant log files
        self._log_data = l2l.read_log_data(self.log_dir)

        # Do a check to make sure the performance measure is logged
        if self.perf_measure not in self._log_data.columns:
            raise KeyError(
                f"Performance measure ({self.perf_measure}) not found in the log data"
            )

        # Validate scenario info
        self.scenario_info = l2l.read_scenario_info(self.log_dir)

        # Validate data format
        l2l.validate_log(self._log_data, self.logger_info["metrics_columns"])

        # Filter data by completed experiences
        self._log_data = self._log_data[self._log_data["exp_status"] == "complete"]

        # Drop all rows with NaN values
        self._log_data = self._log_data[self._log_data[self.perf_measure].notna()]

        # Modify logs for variant-aware or variant-agnostic calculations
        if self.variant_mode == "agnostic":
            # Remove variant label from task names
            self._log_data.task_name = self._log_data.task_name.apply(
                lambda x: x.split("_")[0]
            )

        # Check for log data after filtering
        if self._log_data.empty:
            raise ValueError(
                f"Logs do not contain any valid data for: {self.perf_measure}"
            )

        # Fill in regime number and sort
        self._log_data = l2l.fill_regime_num(self._log_data)
        self._log_data = self._log_data.sort_values(
            by=["regime_num", "exp_num"]
        ).set_index("regime_num", drop=False)

        # Save raw data as separate column
        self._log_data[self.perf_measure + "_raw"] = self._log_data[
            self.perf_measure
        ].to_numpy()

        # Get block summary
        self.block_info = l2l.parse_blocks(
            self._log_data, include_task_params=self.variant_mode == "aware"
        )

        # Store unique task names by order of training
        self._unique_tasks = list(
            self.block_info.sort_values(
                ["block_type", "block_num"], ascending=[False, True]
            )["task_name"].unique()
        )

        # Load all STE data for tasks in log data
        self.load_ste_data()

        # Smooth LL and STE data
        if self.smoothing_method != "none":
            self.smooth_data()

        # Remove outliers
        if self.clamp_outliers:
            self.filter_outliers(quantiles=(0.1, 0.9))

        # Normalize LL and STE data
        if self.normalization_method != "none":
            self.normalize_data()
        else:
            self.normalizer = None

        # Adds default metrics
        self._add_default_metrics()

        # Initialize a results dictionary that can be returned at the end of the calculation step and an internal
        # dictionary that can be passed around for internal calculations
        self.block_info_keys_to_include = [
            "block_num",
            "block_type",
            "block_subtype",
            "task_name",
            "regime_num",
        ]
        if "task_params" in self.block_info.columns:
            if len(self.block_info.task_params.unique()) > 1:
                self.block_info_keys_to_include.append("task_params")

        self._metrics_df = self.block_info[self.block_info_keys_to_include].copy()

    def add(self, metrics_list: Union[Metric, List[Metric]]) -> None:
        if isinstance(metrics_list, list):
            self._metrics.extend(metrics_list)
        else:
            self._metrics.append(metrics_list)

    def _add_default_metrics(self) -> None:
        # Default metrics no matter the syllabus type
        self.add(BlockSaturation(self.perf_measure))
        self.add(TerminalPerformance(self.perf_measure))
        self.add(RecoveryTime(self.perf_measure))
        self.add(PerformanceRecovery(self.perf_measure))
        self.add(PerformanceMaintenance(self.perf_measure, self.maintenance_method))
        self.add(Transfer(self.perf_measure, self.transfer_method))
        self.add(
            STERelativePerf(self.perf_measure, self.ste_data, self.ste_averaging_method)
        )
        self.add(
            SampleEfficiency(
                self.perf_measure, self.ste_data, self.ste_averaging_method
            )
        )

    def add_noise(self, mean: float, std: float) -> None:
        # Add Gaussian noise to log data
        noise = np.random.normal(mean, std, len(self._log_data[self.perf_measure]))
        self._log_data[self.perf_measure] = self._log_data[self.perf_measure] + noise

    def load_ste_data(self) -> None:
        self.ste_data = {}

        for task in self._unique_tasks:
            ste_data = load_ste_data(task)

            # Drop all rows with NaN values
            for idx, ste_data_df in enumerate(ste_data):
                ste_data[idx] = ste_data_df[ste_data_df[self.perf_measure].notna()]

            self.ste_data[task] = ste_data

    def filter_outliers(self, quantiles: Tuple[float, float] = (0.1, 0.9)) -> None:
        # Filter outliers per-task
        for task in self._unique_tasks:
            # Get task data from dataframe
            x = self._log_data[self._log_data["task_name"] == task][
                self.perf_measure
            ].to_numpy()

            # Initialize bounds
            lower_bound = 0
            upper_bound = 100

            if self.data_range:
                lower_bound = self.data_range[task]["min"]
                upper_bound = self.data_range[task]["max"]
            else:
                if self.ste_data.get(task):
                    x_ste = np.concatenate(
                        [
                            ste_data_df[ste_data_df["block_type"] == "train"][
                                self.perf_measure
                            ].to_numpy()
                            for ste_data_df in self.ste_data.get(task)
                        ]
                    )
                    x_comb = np.append(x, x_ste)
                    lower_bound, upper_bound = np.quantile(x_comb, quantiles)
                else:
                    lower_bound, upper_bound = np.quantile(x, quantiles)

            # Filter LL data
            self._log_data.loc[
                self._log_data["task_name"] == task, self.perf_measure
            ] = x.clip(lower_bound, upper_bound)

            # Filter STE data
            for idx, ste_data_df in enumerate(self.ste_data.get(task, [])):
                x = ste_data_df[ste_data_df["task_name"] == task][
                    self.perf_measure
                ].to_numpy()
                self.ste_data[task][idx].loc[
                    ste_data_df["task_name"] == task, self.perf_measure
                ] = x.clip(lower_bound, upper_bound)

        # Save filtered data as separate column
        self._log_data[self.perf_measure + "_filtered"] = self._log_data[
            self.perf_measure
        ].to_numpy()

    def normalize_data(self) -> None:
        # Instantiate normalizer
        self.normalizer = Normalizer(
            perf_measure=self.perf_measure,
            data=self._log_data[["task_name", self.perf_measure]].set_index(
                "task_name"
            ),
            ste_data=self.ste_data,
            data_range=self.data_range,
            method=self.normalization_method,
        )

        # Normalize LL data
        self._log_data = self.normalizer.normalize(self._log_data)

        # Save normalized data as separate column
        self._log_data[self.perf_measure + "_normalized"] = self._log_data[
            self.perf_measure
        ].to_numpy()

        # Normalize STE data
        for task, ste_data in self.ste_data.items():
            if ste_data is not None:
                for idx, ste_data_df in enumerate(ste_data):
                    self.ste_data[task][idx] = self.normalizer.normalize(ste_data_df)

    def smooth_data(self) -> None:
        # Smooth LX data
        for regime_num in self.block_info["regime_num"].unique():
            if (
                self.block_info.iloc[regime_num].block_type == "train"
                or self.do_smooth_eval_data
            ):
                x = self._log_data[self._log_data["regime_num"] == regime_num][
                    self.perf_measure
                ].to_numpy()
                self._log_data.loc[
                    self._log_data["regime_num"] == regime_num, self.perf_measure
                ] = smooth(
                    x, window_len=self.window_length, window=self.smoothing_method
                )

        # Save smoothed data as separate column
        self._log_data[self.perf_measure + "_smoothed"] = self._log_data[
            self.perf_measure
        ].to_numpy()

        # Smooth STE data
        for task, ste_data in self.ste_data.items():
            if ste_data is not None:
                for idx, ste_data_df in enumerate(ste_data):
                    for regime_num in ste_data_df["regime_num"].unique():
                        x = ste_data_df[ste_data_df["regime_num"] == regime_num][
                            self.perf_measure
                        ].to_numpy()
                        self.ste_data[task][idx].loc[
                            ste_data_df["regime_num"] == regime_num, self.perf_measure
                        ] = smooth(
                            x,
                            window_len=self.window_length,
                            window=self.smoothing_method,
                        )

    def log_summary(self) -> pd.DataFrame:
        # Get summary of log data
        task_experiences = {"task_name": [], "LX": [], "EX": []}

        for task in self._unique_tasks:
            task_experiences["task_name"].append(task)
            task_experiences["LX"].append(
                self._log_data[
                    (self._log_data["task_name"] == task)
                    & (self._log_data["block_type"] == "train")
                ].shape[0]
            )
            task_experiences["EX"].append(
                self._log_data[
                    (self._log_data["task_name"] == task)
                    & (self._log_data["block_type"] == "test")
                ].shape[0]
            )

        return pd.DataFrame(task_experiences).set_index("task_name")

    def calculate(self) -> None:
        for metric in self._metrics:
            self._metrics_df = metric.calculate(
                self._log_data, self.block_info, self._metrics_df
            )

        self.calculate_regime_metrics()
        self.calculate_task_metrics()
        self.calculate_lifetime_metrics()

        # Get performance data stats (#LX, #EX for each task)
        log_summary = self.log_summary()
        data_min = np.nanmin(self._log_data[self.perf_measure])
        data_max = np.nanmax(self._log_data[self.perf_measure])
        num_lx = int(log_summary["LX"].sum())
        num_ex = int(log_summary["EX"].sum())

        # Append scenario information to metrics dataframe
        self.ll_metrics_df = self.lifetime_metrics_df.copy()
        if self.ll_metrics_df.empty:
            self.ll_metrics_df = self.ll_metrics_df.append(
                pd.Series([np.nan]), ignore_index=True
            )
        self.regime_metrics_df["run_id"] = Path(self.log_dir).name
        self.ll_metrics_df["run_id"] = Path(self.log_dir).name
        self.ll_metrics_df["complexity"] = self.scenario_info["complexity"]
        self.ll_metrics_df["difficulty"] = self.scenario_info["difficulty"]
        self.ll_metrics_df["scenario_type"] = self.scenario_info["scenario_type"]
        self.ll_metrics_df["metrics_column"] = self.perf_measure
        self.ll_metrics_df["min"] = data_min
        self.ll_metrics_df["max"] = data_max
        self.ll_metrics_df["num_lx"] = num_lx
        self.ll_metrics_df["num_ex"] = num_ex

        timestamps = self._log_data.timestamp.dropna().astype(str)
        max_time = dt.strptime(np.nanmax(timestamps), "%Y%m%dT%H%M%S.%f")
        min_time = dt.strptime(np.nanmin(timestamps), "%Y%m%dT%H%M%S.%f")
        self.ll_metrics_df["runtime"] = (max_time - min_time).total_seconds()

        # Build JSON
        self.ll_metrics_dict = json.loads(self.ll_metrics_df.loc[0].T.to_json())
        self.ll_metrics_dict["normalization_data_range"] = (
            self.normalizer.data_range if self.normalizer else None
        )
        self.ll_metrics_dict["task_metrics"] = self.task_metrics_df.T.to_dict()

        for task in self._unique_tasks:
            self.ll_metrics_dict["task_metrics"][task]["min"] = np.nanmin(
                self._log_data[self._log_data["task_name"] == task][self.perf_measure]
            )
            self.ll_metrics_dict["task_metrics"][task]["max"] = np.nanmax(
                self._log_data[self._log_data["task_name"] == task][self.perf_measure]
            )
            self.ll_metrics_dict["task_metrics"][task]["num_lx"] = int(
                log_summary.loc[task, "LX"]
            )
            self.ll_metrics_dict["task_metrics"][task]["num_ex"] = int(
                log_summary.loc[task, "EX"]
            )

    def calculate_regime_metrics(self) -> None:
        # Create dataframe for regime-level metrics
        regime_metrics = ["saturation", "exp_to_sat", "term_perf", "exp_to_term_perf"]
        self.regime_metrics_df = self.block_info[self.block_info_keys_to_include]

        # Fill regime metrics dataframe
        self.regime_metrics_df = pd.concat(
            [self.regime_metrics_df, self._metrics_df[regime_metrics]], axis=1
        )
        if "task_params" in self.block_info_keys_to_include:
            if self.regime_metrics_df["task_params"].size:
                self.regime_metrics_df["task_params"] = (
                    self.regime_metrics_df["task_params"]
                    .dropna()
                    .apply(lambda x: x[:25] + "..." if len(x) > 25 else x)
                )
            else:
                self.regime_metrics_df = self.regime_metrics_df.dropna(axis=1)

    def calculate_task_metrics(self) -> None:
        # Initialize task metrics dataframe
        self.task_metrics_df = pd.DataFrame(
            index=self._unique_tasks, columns=self.task_metrics
        )
        self.task_metrics_df.index.name = "task_name"

        # Initialize certain task metrics data objects
        num_tasks = len(self._unique_tasks)
        self.task_metrics_df["recovery_times"] = [[]] * num_tasks
        if self.maintenance_method in ["mrlep", "both"]:
            self.task_metrics_df["maintenance_val_mrlep"] = [[]] * num_tasks
        if self.maintenance_method in ["mrtlp", "both"]:
            self.task_metrics_df["maintenance_val_mrtlp"] = [[]] * num_tasks
        if self.transfer_method in ["ratio", "both"]:
            self.task_metrics_df["forward_transfer_ratio"] = [{}] * num_tasks
            self.task_metrics_df["backward_transfer_ratio"] = [{}] * num_tasks
            self.forward_transfer_ratio = defaultdict(dict)
            self.backward_transfer_ratio = defaultdict(dict)
        if self.transfer_method in ["contrast", "both"]:
            self.task_metrics_df["forward_transfer_contrast"] = [{}] * num_tasks
            self.task_metrics_df["backward_transfer_contrast"] = [{}] * num_tasks
            self.forward_transfer_contrast = defaultdict(dict)
            self.backward_transfer_contrast = defaultdict(dict)
        self.task_metrics_df["ste_rel_perf_vals"] = [[]] * num_tasks
        self.task_metrics_df["ste_saturation_vals"] = [[]] * num_tasks
        self.task_metrics_df["ste_exp_to_sat_vals"] = [[]] * num_tasks
        self.task_metrics_df["se_saturation_vals"] = [[]] * num_tasks
        self.task_metrics_df["se_exp_to_sat_vals"] = [[]] * num_tasks
        self.task_metrics_df["sample_efficiency_vals"] = [[]] * num_tasks

        # Create data structures for transfer values
        for _, row in self._metrics_df.iterrows():
            for transfer_metric in [
                "forward_transfer_contrast",
                "forward_transfer_ratio",
                "backward_transfer_contrast",
                "backward_transfer_ratio",
            ]:
                if (
                    transfer_metric in self._metrics_df
                    and type(row[transfer_metric]) is dict
                ):
                    [(other_task, transfer_value)] = row[transfer_metric].items()
                    key = (other_task, row["task_name"])
                    if key[0] not in getattr(self, transfer_metric).keys():
                        getattr(self, transfer_metric)[key[0]][key[1]] = [
                            transfer_value
                        ]
                    elif key[1] not in getattr(self, transfer_metric)[key[0]].keys():
                        getattr(self, transfer_metric)[key[0]][key[1]] = [
                            transfer_value
                        ]
                    else:
                        getattr(self, transfer_metric)[key[0]][key[1]].append(
                            transfer_value
                        )

        # Fill task metrics dataframe
        for task in self._unique_tasks:
            # Get task metrics
            tm = self._metrics_df[self._metrics_df["task_name"] == task]

            # Iterate over task metrics
            for metric in self.task_metrics:
                if metric in tm.keys():
                    if metric == "perf_recovery":
                        pr = tm[metric].dropna().to_numpy(dtype=float)
                        self.task_metrics_df.at[task, metric] = (
                            np.NaN if len(pr) == 0 else pr[0]
                        )
                        self.task_metrics_df.at[task, "recovery_times"] = list(
                            tm["recovery_time"].dropna().to_numpy(dtype=float)
                        )
                    elif metric in ["perf_maintenance_mrtlp", "perf_maintenance_mrlep"]:
                        pm = tm[metric].dropna().to_numpy(dtype=float)
                        self.task_metrics_df.at[task, metric] = (
                            pm[0] if len(pm) else np.NaN
                        )
                        maintenance_val_name = (
                            "maintenance_val_" + metric.split("_")[-1]
                        )
                        maintenance_values = list(
                            tm[maintenance_val_name].to_numpy(dtype=float)
                        )
                        self.task_metrics_df.at[task, maintenance_val_name] = [
                            maintenance_values[s]
                            for s in np.ma.clump_unmasked(
                                np.ma.masked_invalid(maintenance_values)
                            )
                        ]
                    elif metric in [
                        "forward_transfer_contrast",
                        "forward_transfer_ratio",
                        "backward_transfer_contrast",
                        "backward_transfer_ratio",
                    ]:
                        self.task_metrics_df.at[task, metric] = getattr(self, metric)[
                            task
                        ]
                    elif metric == "ste_rel_perf":
                        rp = tm[metric].dropna().to_numpy()
                        if rp.size == 0:
                            self.task_metrics_df.at[task, "ste_rel_perf_vals"] = []
                            self.task_metrics_df.at[task, metric] = np.NaN
                        elif rp.size == 1:
                            self.task_metrics_df.at[task, "ste_rel_perf_vals"] = rp[0]
                            self.task_metrics_df.at[task, metric] = np.nanmean(rp[0])
                        else:
                            raise ValueError("Unexpected size for relative performance")
                    elif metric == "sample_efficiency":
                        task_sat = (
                            tm["se_task_saturation"].dropna().to_numpy(dtype=float)
                        )
                        task_exp_to_sat = (
                            tm["se_task_exp_to_sat"].dropna().to_numpy(dtype=float)
                        )
                        ste_saturation = tm["se_ste_saturation"].dropna().to_numpy()
                        ste_exp_to_sat = tm["se_ste_exp_to_sat"].dropna().to_numpy()
                        se_saturation = tm["se_saturation"].dropna().to_numpy()
                        se_exp_to_sat = tm["se_exp_to_sat"].dropna().to_numpy()
                        se = tm[metric].dropna().to_numpy()

                        if se.size == 0:
                            self.task_metrics_df.at[task, "se_task_saturation"] = np.NaN
                            self.task_metrics_df.at[task, "se_task_exp_to_sat"] = np.NaN
                            self.task_metrics_df.at[task, "ste_saturation_vals"] = []
                            self.task_metrics_df.at[task, "ste_exp_to_sat_vals"] = []
                            self.task_metrics_df.at[task, "se_saturation_vals"] = []
                            self.task_metrics_df.at[task, "se_exp_to_sat_vals"] = []
                            self.task_metrics_df.at[task, "sample_efficiency_vals"] = []
                            self.task_metrics_df.at[task, metric] = np.NaN
                        elif se.size == 1:
                            self.task_metrics_df.at[
                                task, "se_task_saturation"
                            ] = task_sat[0]
                            self.task_metrics_df.at[
                                task, "se_task_exp_to_sat"
                            ] = task_exp_to_sat[0]
                            self.task_metrics_df.at[
                                task, "ste_saturation_vals"
                            ] = ste_saturation[0]
                            self.task_metrics_df.at[
                                task, "ste_exp_to_sat_vals"
                            ] = ste_exp_to_sat[0]
                            self.task_metrics_df.at[
                                task, "se_saturation_vals"
                            ] = se_saturation[0]
                            self.task_metrics_df.at[
                                task, "se_exp_to_sat_vals"
                            ] = se_exp_to_sat[0]
                            self.task_metrics_df.at[
                                task, "sample_efficiency_vals"
                            ] = se[0]
                            self.task_metrics_df.at[task, metric] = np.nanmean(se[0])
                        else:
                            raise ValueError("Unexpected size for sample efficiency")
                    else:
                        # Drop NaN values
                        metric_values = tm[metric].dropna().to_numpy(dtype=float)

                        if len(metric_values) == 0:
                            self.task_metrics_df.at[task, metric] = np.NaN
                        elif len(metric_values) == 1:
                            self.task_metrics_df.at[task, metric] = metric_values[0]
                        else:
                            self.task_metrics_df.at[task, metric] = metric_values

    def calculate_lifetime_metrics(self) -> None:
        # Calculate lifetime metrics from task metrics
        self.lifetime_metrics_df = pd.DataFrame(columns=self.task_metrics)

        for metric in self.task_metrics:
            if metric in self.task_metrics_df:
                if metric in [
                    "forward_transfer_contrast",
                    "forward_transfer_ratio",
                    "backward_transfer_contrast",
                    "backward_transfer_ratio",
                ]:
                    # Get the first calculated transfer values for each task pair
                    metric_vals = np.array(
                        [
                            v2[0]
                            for _, v in getattr(self, metric).items()
                            for _, v2 in v.items()
                        ]
                    )
                else:
                    metric_vals = (
                        self.task_metrics_df[metric].dropna().to_numpy(dtype=float)
                    )

                if len(metric_vals):
                    # Aggregate metric values
                    if self.aggregation_method == "mean":
                        self.lifetime_metrics_df[metric] = [
                            np.nanmean(metric_vals.astype(np.float))
                        ]
                    elif self.aggregation_method == "median":
                        self.lifetime_metrics_df[metric] = [
                            np.nanmedian(metric_vals.astype(np.float))
                        ]

    def report(self) -> None:
        """Print summary report of lifetime metrics and return metric objects."""
        # TODO: Handle reporting custom metrics
        # Print lifetime metrics
        print("\nLifetime Metrics:")
        print(
            tabulate(
                self.lifetime_metrics_df.fillna("N/A"),
                headers="keys",
                tablefmt="psql",
                floatfmt=".2f",
                showindex=False,
            )
        )

    def save_metrics(self, output_dir: str = "", filename: str = None):
        """Save metrics out as JSON file.

        Args:
            output_dir (str, optional): Output directory. Defaults to ''.
            filename (str, optional): Base filename for metrics file. Defaults to log directory name.
        """

        # Generate filename
        if filename is None:
            filename = Path(self.log_dir).name
        else:
            filename = filename.replace(" ", "_")

        # Save metrics to file
        with open(
            Path(output_dir) / (filename + "_metrics.json"), "w", newline="\n"
        ) as metrics_file:
            json.dump(self.ll_metrics_dict, metrics_file)
        self.regime_metrics_df.to_csv(
            Path(output_dir) / (filename + "_regime_metrics.tsv"), sep="\t"
        )

    def save_data(self, output_dir: str = "", filename: str = None) -> None:
        """Save out raw and processed data.

        Args:
            output_dir (str, optional): Output directory. Defaults to ''.
            filename (str, optional): Base filename for data file. Defaults to log directory name.
        """

        # Generate filename
        if filename is None:
            filename = Path(self.log_dir).name
        else:
            filename = filename.replace(" ", "_")

        # Save data
        self._log_data.reset_index(drop=True).to_feather(
            str(Path(output_dir) / (filename + "_data.feather"))
        )

    def save_settings(self, output_dir: str = "", filename: str = None) -> None:
        """Save out settings used to calculate metrics.

        Args:
            output_dir (str, optional): Output directory. Defaults to ''.
            filename (str, optional): Base filename for settings file. Defaults to log directory name.
        """
        # Generate filename
        if filename is None:
            filename = Path(self.log_dir).name
        else:
            filename = filename.replace(" ", "_")

        # Build settings JSON
        settings_json = {}
        settings_json["log_dir"] = str(self.log_dir)
        settings_json["perf_measure"] = self.perf_measure
        settings_json["variant_mode"] = self.variant_mode
        settings_json["ste_averaging_method"] = self.ste_averaging_method
        settings_json["aggregation_method"] = self.aggregation_method
        settings_json["maintenance_method"] = self.maintenance_method
        settings_json["transfer_method"] = self.transfer_method
        settings_json["normalization_method"] = self.normalization_method
        settings_json["smoothing_method"] = self.smoothing_method
        settings_json["window_length"] = self.window_length
        settings_json["clamp_outliers"] = self.clamp_outliers

        with open(Path(output_dir) / (filename + "_settings.json"), "w") as outfile:
            json.dump(settings_json, outfile)

    def plot(
        self,
        save: bool = False,
        show_eval_lines: bool = True,
        output_dir: str = "",
        task_colors: dict = {},
        input_title: str = None,
    ) -> None:

        if input_title is None:
            input_title = Path(self.log_dir).name
        plot_filename = input_title

        plot_blocks(
            self._log_data,
            self.perf_measure,
            self._unique_tasks,
            task_colors=task_colors,
            input_title=input_title,
            output_dir=output_dir,
            do_save_fig=save,
            plot_filename=plot_filename + "_block",
        )
        plot_performance(
            self._log_data,
            self.block_info,
            unique_tasks=self._unique_tasks,
            task_colors=task_colors,
            show_eval_lines=show_eval_lines,
            y_axis_col=self.perf_measure,
            input_title=input_title,
            output_dir=output_dir,
            do_save_fig=save,
            plot_filename=plot_filename + "_perf",
        )

    def plot_ste_data(
        self,
        input_title: str = None,
        save: bool = False,
        output_dir: str = "",
        task_colors: dict = {},
    ) -> None:
        if input_title is None:
            input_title = "Performance Relative to STE\n" + Path(self.log_dir).name
        plot_filename = Path(self.log_dir).name + "_ste"

        plot_ste_data(
            self._log_data,
            self.ste_data,
            self.block_info,
            self._unique_tasks,
            task_colors=task_colors,
            perf_measure=self.perf_measure,
            ste_averaging_method=self.ste_averaging_method,
            input_title=input_title,
            output_dir=output_dir,
            do_save=save,
            plot_filename=plot_filename,
        )



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging
from collections import defaultdict
from typing import Tuple

import pandas as pd

from ._localutil import fill_metrics_df, get_terminal_perf
from .core import Metric

logger = logging.getLogger(__name__)


class RecoveryTime(Metric):
    name = "Recovery Time"
    capability = "adapt_to_new_tasks"
    requires = {"syllabus_type": "agent"}
    description = "Calculates whether the system recovers after a change of task or parameters and \
        calculate how long it takes if recovery is achieved"

    def __init__(self, perf_measure: str) -> None:
        super().__init__()
        self.perf_measure = perf_measure

    def validate(self, block_info: pd.DataFrame) -> Tuple[dict, dict]:
        # Get unique tasks
        unique_tasks = block_info.loc[:, "task_name"].unique()

        # Determine where we need to assess recovery time for each task
        ref_indices = defaultdict(list)
        assess_indices = defaultdict(list)

        for task in unique_tasks:
            # Get train blocks in order of appearance
            tr_block_info = block_info.sort_index().loc[
                (block_info["block_type"] == "train")
                & (block_info["block_subtype"] == "wake")
                & (block_info["task_name"] == task),
                ["task_name"],
            ]
            tr_indices = tr_block_info.index

            # Regimes are defined as new combinations of tasks and params, but can repeat,
            # so check for changes across regimes
            first = True
            for idx, block_idx in enumerate(tr_indices):
                if first:
                    first = False
                    continue
                assess_indices[task].append(block_idx)
                ref_indices[task].append(tr_indices[idx - 1])

        if not (ref_indices and assess_indices):
            raise ValueError("Not enough blocks to assess recovery time")

        return ref_indices, assess_indices

    def calculate(
        self,
        dataframe: pd.DataFrame,
        block_info: pd.DataFrame,
        metrics_df: pd.DataFrame,
    ) -> pd.DataFrame:
        try:
            # Get the places where we should calculate recovery time
            ref_indices, assess_indices = self.validate(block_info)

            # Initialize metric dictionary
            recovery_time = {}

            # Iterate over indices for computing recovery time
            for (_, ref_vals), (_, assess_vals) in zip(
                ref_indices.items(), assess_indices.items()
            ):
                for ref_ind, assess_ind in zip(ref_vals, assess_vals):
                    prev_val = metrics_df["term_perf"][ref_ind]
                    block_data = dataframe.loc[assess_ind]

                    # Check for proper number of rows in block data
                    if block_data.ndim == 1:
                        block_data = pd.DataFrame(block_data).T

                    _, _, exp_to_rec = get_terminal_perf(
                        block_data, col_to_use=self.perf_measure, prev_val=prev_val
                    )
                    recovery_time[assess_ind] = exp_to_rec

            return fill_metrics_df(recovery_time, "recovery_time", metrics_df)
        except ValueError as e:
            logger.warning(f"Cannot compute {self.name} - {e}")
            return metrics_df



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging

import numpy as np
import pandas as pd
from scipy import stats

from ._localutil import fill_metrics_df
from .core import Metric

logger = logging.getLogger(__name__)


class PerformanceRecovery(Metric):
    name = "Performance Recovery"
    capability = "adapt_to_new_tasks"
    requires = {"syllabus_type": "agent"}
    description = "Calculates the performance recovery value corresponding to a change of task or parameters"

    def __init__(self, perf_measure: str) -> None:
        super().__init__()
        self.perf_measure = perf_measure

    def validate(self, metrics_df: pd.DataFrame) -> None:
        # Get number of recovery times
        if "recovery_time" in metrics_df.columns:
            r = metrics_df["recovery_time"]
            r = r[r.notna()]
            r_count = r.count()

            if r_count <= 1:
                raise ValueError(
                    "Not enough recovery times to assess performance recovery"
                )
        else:
            raise ValueError("No recovery times")

    def calculate(
        self,
        dataframe: pd.DataFrame,
        block_info: pd.DataFrame,
        metrics_df: pd.DataFrame,
    ) -> pd.DataFrame:
        try:
            # Get the places where we should calculate recovery time
            self.validate(metrics_df)

            # Initialize metric dictionary
            pr_values = {}

            # Calculate performance recovery for each task
            for task in block_info.loc[:, "task_name"].unique():
                r = metrics_df[metrics_df["task_name"] == task]["recovery_time"]
                r = r[r.notna()]

                # Get Theil-Sen slope
                y = np.array(r)

                if len(y) > 1:
                    slope, _, _, _ = stats.theilslopes(y)

                    # Set performance recovery value as negative slope (greater is better)
                    idx = block_info[block_info["task_name"] == task][
                        "regime_num"
                    ].max()
                    pr_values[idx] = -slope
                else:
                    logger.warning(
                        f"Cannot compute {self.name} for task {task} - Not enough recovery times"
                    )

            return fill_metrics_df(pr_values, "perf_recovery", metrics_df)
        except ValueError as e:
            logger.warning(f"Cannot compute {self.name} - {e}")
            return metrics_df



"""
Copyright © 2021-2022 The Johns Hopkins University Applied Physics Laboratory LLC

Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the “Software”), to 
deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is 
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in 
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR 
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import logging
from collections import defaultdict
from typing import Set

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


class Normalizer:
    """Utility class for normalizing data."""

    valid_methods = ["task", "run"]

    def __init__(
        self,
        perf_measure: str,
        data: pd.DataFrame,
        ste_data: dict = {},
        data_range: defaultdict = None,
        method: str = "task",
        scale: int = 100,
        offset: int = 1,
    ) -> None:
        """Constructor for Normalizer.

        Args:
            perf_measure (str): Name of column to use for metrics calculations.
            data (pd.DataFrame, optional): Reference data for calculating data range. Assumed
                DataFrame with task name as index and one column of performance data.
            ste_data (dict, optional): The STE data for computing quantiles. Defaults to {}.
            data_range (defaultdict, optional): Dictionary object for data range. Defaults to None.
            method (str, optional): Normalization method. Valid values are 'task' and 'run'.
                Defaults to 'task'.
            scale (int, optional): Normalization scale, interpreted as from 0 to this value.
                Defaults to 100.
            offset (int, optional): Offset to normalized data. Defaults to 1.

        Raises:
            ValueError: If data range validation fails.
        """

        self.perf_measure = perf_measure

        # Get unique task names in data
        self.unique_tasks = set(data.index.unique())

        if data_range is not None:
            # Validate and set data range for normalizer
            if self._validate_data_range(
                data_range=data_range, task_names=self.unique_tasks
            ):
                self.data_range = data_range
                self.run_min = min([val["min"] for val in self.data_range.values()])
                self.run_max = max([val["max"] for val in self.data_range.values()])
        elif data is not None:
            self.calculate_data_range(data, ste_data)
        else:
            raise ValueError(
                f"Must provide data or data range to initialize Normalizer"
            )

        if self._validate_method(method):
            self.method = method

        if self._validate_scale(scale):
            self.scale = scale

        if self._validate_offset(offset):
            self.offset = offset

    def calculate_data_range(self, data: pd.DataFrame, ste_data: dict = {}) -> None:
        """Calculates data range per task for given data.

        A task data range is the minimum and maximum value of the task performance.

        Args:
            data (pd.DataFrame): Reference data for calculating data range. Assumed
                DataFrame with task name as index and one column of performance data.
            ste_data (dict, optional): The STE data for computing quantiles. Defaults to {}.

        Raises:
            ValueError: If data contains more than just performance values and task name.
        """

        data_column = data.columns.to_numpy()

        if len(data_column) > 1:
            raise ValueError(
                f"Data must only have one column with performance measures"
            )

        # Initialize data range as empty object
        self.data_range = defaultdict(dict)

        # Get data range over scenario and STE data if not provided as input
        for task in self.unique_tasks:
            # Get feature range for each task
            task_min = np.nanmin(data.loc[task])
            task_max = np.nanmax(data.loc[task])

            if ste_data.get(task):
                x_ste = np.concatenate(
                    [
                        ste_data_df[ste_data_df["block_type"] == "train"][
                            self.perf_measure
                        ].to_numpy()
                        for ste_data_df in ste_data.get(task)
                    ]
                )
                self.data_range[task]["min"] = min(task_min, np.nanmin(x_ste))
                self.data_range[task]["max"] = max(task_max, np.nanmax(x_ste))
            else:
                self.data_range[task]["min"] = task_min
                self.data_range[task]["max"] = task_max

        self.run_min = min([val["min"] for val in self.data_range.values()])
        self.run_max = max([val["max"] for val in self.data_range.values()])

    def _validate_data_range(
        self, data_range: defaultdict, task_names: Set[str]
    ) -> bool:
        """Validates data range object.

        Args:
            data_range (defaultdict): Dictionary object for data range.
            task_names (Set[str]): Set of task names in the data.

        Raises:
            TypeError: If data range is not a dictionary object.
            KeyError: If data range is not defined for all tasks.
            KeyError: If the keys min and max are missing.

        Returns:
            bool: True if validation succeeds.
        """

        if not isinstance(data_range, (dict, defaultdict)):
            raise TypeError(f"Invalid data range type - Must be a dictionary")
        elif not set(data_range.keys()).issuperset(task_names):
            raise KeyError(f"Data range not defined for all tasks: {task_names}")
        elif False in [key.keys() >= {"min", "max"} for key in data_range.values()]:
            raise KeyError(f"Missing required fields: min and max")
        else:
            return True

    def _validate_method(self, method: str) -> bool:
        """Validates normalization method.

        Args:
            method (str): Normalization method.

        Raises:
            ValueError: If method is not in list of valid methods.

        Returns:
            bool: True if validation succeeds.
        """

        if method not in self.valid_methods:
            raise ValueError(
                f"Invalid normalization method: {method}\n"
                f"Valid methods are: {self.valid_methods}"
            )
        else:
            return True

    def _validate_scale(self, scale: int) -> bool:
        """Validates normalization scale.

        Args:
            scale (int): Normalization scale.

        Raises:
            TypeError: If scale is not an integer.
            ValueError: If scale is less than or equal to 0.

        Returns:
            bool: True if validation succeeds.
        """

        if not isinstance(scale, int):
            raise TypeError(f"Invalid scale type: {type(scale)}")
        elif scale <= 0:
            raise ValueError(f"Scale value must be greater than 0: {scale}")
        else:
            return True

    def _validate_offset(self, offset: int) -> bool:
        """Validates normalization offset.

        Args:
            offset (int): Normalization offset.

        Raises:
            TypeError: If offset is not an integer.
            ValueError: If scale is less than or equal to 0.

        Returns:
            bool: True if validation succeeds.
        """

        if not isinstance(offset, int):
            raise TypeError(f"Invalid offset type: {type(offset)}")
        elif offset <= 0:
            raise ValueError(f"Offset value must be greater than 0: {offset}")
        else:
            return True

    def normalize(self, data: pd.DataFrame) -> pd.DataFrame:
        """Normalizes the given data with the current instance method and data range/scale.

        Args:
            data (pd.DataFrame): Dataframe to be normalized.

        Raises:
            KeyError: If there's a missing data range for any task.

        Returns:
            pd.DataFrame: Normalized dataframe.
        """

        if self.method == "task":
            for task in data["task_name"].unique():
                if task in self.data_range.keys():
                    task_min = self.data_range[task]["min"]
                    task_max = self.data_range[task]["max"]
                    task_data = data.loc[
                        data["task_name"] == task, self.perf_measure
                    ].to_numpy()

                    if task_min == task_max:
                        data.loc[data["task_name"] == task, self.perf_measure] = (
                            task_data * 0
                        ) + self.offset
                        logger.warning(
                            f"Performance for task ({task}) is constant - normalizing to 0"
                        )
                    else:
                        data.loc[data["task_name"] == task, self.perf_measure] = (
                            (task_data - task_min) / (task_max - task_min) * self.scale
                        ) + self.offset
                else:
                    raise KeyError(f"Missing data range for task '{task}'")
            return data
        elif self.method == "run":
            data.loc[:, self.perf_measure] = (
                (data[self.perf_measure].to_numpy() - self.run_min)
                / (self.run_max - self.run_min)
                * self.scale
            ) + self.offset
            return data
